import datetime
import os
import subprocess

import pytest

from vcsinfo.vcsinfo import Repository
from vcsinfo.factory import build
from vcsinfo.exceptions import InvalidInputError, VcsError
from vcsinfo.implementations.git import GitRepository, GitVersion


@pytest.fixture(scope="module")
def repo(tmpdir_factory) -> GitRepository:
    repo_dir = str(tmpdir_factory.mktemp('repo'))
    return make_repo(repo_dir)


@pytest.fixture(scope="module")
def other_repo(tmpdir_factory) -> GitRepository:
    repo_dir = str(tmpdir_factory.mktemp('other_repo'))
    return make_repo(repo_dir)


@pytest.fixture(scope="module")
def singlecommit_repo(tmpdir_factory) -> GitRepository:
    repo_dir = str(tmpdir_factory.mktemp('singlecommit_repo'))
    return make_repo(repo_dir, 'singlecommit.git')


def make_repo(path: str, bare='simplerepo.git') -> GitRepository:
    bare_repo_dir = os.path.join(os.path.dirname(__file__), 'data', bare)
    subprocess.run(['git', 'clone', bare_repo_dir, path])
    return GitRepository(path)


def test_no_git(repo: Repository) -> None:
    old_environ = os.environ.copy()
    current_environ = os.environ
    current_environ['PATH'] = ''
    assert not GitRepository.is_git_installed()
    with pytest.raises(VcsError):
        GitRepository(repo.path)
    os.environ = old_environ


def test_factory(repo: Repository) -> None:
    built_repo = build(repo.path)
    assert repo == built_repo
    assert repo.path == built_repo.path


def test_factory_fail(tmpdir) -> None:
    with pytest.raises(InvalidInputError):
        build(str(tmpdir))
    with pytest.raises(InvalidInputError):
        build(os.path.join(str(tmpdir), 'thissubdirdoesnotexist'))


def test_repo(repo: Repository) -> None:
    assert repo.current_version().id == '85bf86a8c6747ef4ec11299d1aa2c281dd05e168'


def test_remotes(repo: GitRepository) -> None:
    assert list(repo.remotes) == ['origin']


def test_branches(repo: GitRepository) -> None:
    assert list(repo.branches) == ['v1.0', 'v1.1', 'master', 'origin/HEAD', 'origin/master', 'origin/featurebranch']


def test_is_gitdir(repo: GitRepository) -> None:
    assert GitRepository.is_compatible_path(repo.path)


def test_is_not_gitdir(tmpdir) -> None:
    assert not GitRepository.is_compatible_path(str(tmpdir))
    with pytest.raises(InvalidInputError):
        GitRepository(str(tmpdir))


@pytest.mark.parametrize(
    'ref, sha1',
    [
        ('v1.0', 'd3017a912193acd35facd657c65336413ce4b1e2'),
        ('v1.1', 'eea5e0305701e60f5be6c34e6e54627ecd64b353'),
        ('master', '85bf86a8c6747ef4ec11299d1aa2c281dd05e168'),
        ('origin/master', '85bf86a8c6747ef4ec11299d1aa2c281dd05e168'),
        ('origin/featurebranch', 'e779089202354d3548b0d56284d164e760acf0f6')
    ]
)
def test_refs(repo: GitRepository, ref, sha1) -> None:
    v1 = GitVersion(repo, ref)
    v2 = GitVersion(repo, sha1)
    assert v1.id == sha1
    assert v1 == v2


@pytest.mark.parametrize(
    'ref, timestamp',
    [
        ('v1.0', 1483972814),
        ('v1.1', 1483973037),
        ('master', 1483973259),
        ('origin/master', 1483973259),
        ('origin/featurebranch', 1483973320)
    ]
)
def test_version_date(repo: GitRepository, ref, timestamp) -> None:
    version = GitVersion(repo, ref)
    date = datetime.datetime.fromtimestamp(timestamp)
    assert version.date == date
    assert version.date < datetime.datetime.now()


@pytest.mark.parametrize(
    'ref1, ref2, is_greater',
    [
        ('v1.0', 'v1.0', False),
        ('v1.1', 'v1.0', True),
        ('master', 'v1.0', True)
    ]
)
def test_version_order(repo: GitRepository, other_repo: GitRepository, ref1, ref2, is_greater) -> None:
    v1 = GitVersion(repo, ref1)
    v2 = GitVersion(repo, ref2)
    v3 = GitVersion(other_repo, ref1)
    assert (v1 > v2) == is_greater
    with pytest.raises(TypeError):
        v1 > v3


@pytest.mark.parametrize(
    'rev, branch, has_newer',
    [
        ('v1.0', 'v1.*', True),
        ('v1.0', 'v1.0', False),
        ('v1.1', 'master', True),
        ('v1.0', 'master', True),
        ('master', 'master', False),
        ('origin/featurebranch', 'origin/featurebranch', False),
        ('master', 'origin/featurebranch', False),
        ('master', 'doesnotexist', False)
    ]
)
def test_exists_newer(repo: GitRepository, rev, branch, has_newer) -> None:
    version = GitVersion(repo, rev)
    assert (version.newest(branch=branch).date > version.date) == has_newer


@pytest.mark.parametrize(
    'rev, file, branch, has_newer',
    [
        ('v1.0', 'file.txt', None, True),
        ('master', 'file.txt', None, False),
        ('master', 'file.txt', 'master', False),
        ('v1.0', 'doesnotexist', None, False)
    ]
)
def test_exists_newer_file(repo: GitRepository, rev, file, branch, has_newer) -> None:
    version = GitVersion(repo, rev)
    newest = version.newest(branch=branch).last_change_on(file)
    if newest is None:
        assert not has_newer
    else:
        assert (newest.date > version.date) == has_newer


@pytest.mark.parametrize(
    'rev, branch, newest_sha1',
    [
        ('v1.0', 'v1.*', 'eea5e0305701e60f5be6c34e6e54627ecd64b353'),
        ('v1.0', 'v1.0', 'd3017a912193acd35facd657c65336413ce4b1e2'),
        ('master', 'master', '85bf86a8c6747ef4ec11299d1aa2c281dd05e168'),
        ('origin/featurebranch', 'origin/featurebranch', 'e779089202354d3548b0d56284d164e760acf0f6'),
        ('origin/featurebranch', 'master', 'e779089202354d3548b0d56284d164e760acf0f6'),
        ('origin/featurebranch', 'doesnotexist', 'e779089202354d3548b0d56284d164e760acf0f6'),
        ('master', 'origin/featurebranch', '85bf86a8c6747ef4ec11299d1aa2c281dd05e168'),
        ('master', 'v1.0', '85bf86a8c6747ef4ec11299d1aa2c281dd05e168'),
        ('master', 'doesnotexist', '85bf86a8c6747ef4ec11299d1aa2c281dd05e168')
    ]
)
def test_newest_version(repo: GitRepository, rev, branch, newest_sha1):
    version = GitVersion(repo, rev)
    assert version.newest(branch=branch).id == newest_sha1


@pytest.mark.parametrize(
    'rev, file, branch, newest_sha1',
    [
        ('v1.0', 'file.txt', None, 'e779089202354d3548b0d56284d164e760acf0f6'),
        ('v1.0', 'file.txt', 'master', 'a939b8b474f02fa31cf54ef17cd25f0bba7db90d'),
        ('v1.0', 'anotherfile.txt', 'master', '85bf86a8c6747ef4ec11299d1aa2c281dd05e168'),
        ('v1.0', 'file.txt', 'v1.*', 'eea5e0305701e60f5be6c34e6e54627ecd64b353'),
        ('v1.0', 'anotherfile.txt', 'v1.*', None),
        ('v1.0', 'doesnotexist', None, None)
    ]
)
def test_newest_version_file(repo: GitRepository, rev, file, branch, newest_sha1):
    version = GitVersion(repo, rev)
    newest = version.newest(branch=branch).last_change_on(file)
    if newest is None:
        assert newest_sha1 is None
    else:
        assert newest.id == newest_sha1


@pytest.mark.parametrize(
    'commit1, commit2, is_ancestor',
    [
        ('v1.0', 'v1.1', True),
        ('v1.1', 'v1.0', False),
        ('v1.1', 'master', True),
        ('master', 'v1.1', False),
        ('v1.0', 'master', True),
        ('master', 'v1.0', False),
        ('master', 'origin/featurebranch', False),
        ('origin/featurebranch', 'master', False)
    ]
)
def test_ancestry(
    repo: GitRepository,
    other_repo:
    GitRepository,
    commit1: str,
    commit2: str,
    is_ancestor: bool
) -> None:
    v1 = GitVersion(repo, commit1)
    v2 = GitVersion(repo, commit2)
    v3 = GitVersion(other_repo, commit1)
    assert v1.is_ancestor_of(v2) == is_ancestor
    assert not v1.is_ancestor_of(v3)


@pytest.mark.parametrize(
    'ref',
    [
        'doesnotexist',
        'also doesnt exist'
    ]
)
def test_nonexisting_refs(repo: GitRepository, ref) -> None:
    with pytest.raises(InvalidInputError):
        GitVersion(repo, ref)


@pytest.mark.parametrize(
    'ref, version_id',
    [
        ('v1.0', '1.0'),
        ('v1.1', '1.1'),
        ('master', '1.1.post0.dev2+g85bf86a.master'),
        ('origin/featurebranch', '1.1.post0.dev1+ge779089.origin.featurebranch')
    ]
)
def test_version_id(repo: GitRepository, ref, version_id) -> None:
    version = GitVersion(repo, ref)
    assert version.version_id == version_id


def test_singlecommit_version(singlecommit_repo: GitRepository) -> None:
    assert singlecommit_repo.current_version().version_id == '0.0.0.post0.dev1+g0955b9b.master'
