"""Git implementation of the VCS Info interfaces."""

import datetime
import fnmatch
import functools
import os.path
import re
import typing
import subprocess

import pkg_resources

from ..vcsinfo import Repository, Version
from ..exceptions import InvalidInputError, VcsError


class GitRepository(Repository):
    """Git implementation of a VCS repository."""

    required_git_version = '1.8.5'

    def __init__(self, path: str) -> None:

        if not self.is_git_installed():
            msg = 'The installed version of git (if any) is too old ({required} is required)'
            raise VcsError(msg.format(required=GitRepository.required_git_version))

        if not GitRepository.is_compatible_path(path):
            raise InvalidInputError('{path} is not a git repository'.format(path=path))

        super().__init__(path)
        self.__path = os.path.abspath(path)

    @property
    def path(self) -> str:
        return self.__path

    def current_version(self) -> 'GitVersion':
        return GitVersion(self, 'HEAD')

    @classmethod
    def is_compatible_path(cls, path: str) -> bool:
        return subprocess.run(['git', '-C', path, 'rev-parse']).returncode == 0

    @classmethod
    def is_git_installed(cls) -> bool:
        """Check whether git is installed."""
        try:
            output = subprocess.check_output(['git', '--version']).decode().strip()
            git_version = re.search(r'([0-9]+\.)+[0-9]+', output).group(0)
            return pkg_resources.parse_version(cls.required_git_version) <= pkg_resources.parse_version(git_version)
        except (IndexError, FileNotFoundError, subprocess.CalledProcessError):
            return False

    @property
    def branches(self) -> typing.Iterable[str]:
        """Return all references (branches, tags) of this repository."""
        self.fetch()
        return self.__run_git_command([
            'for-each-ref',
            '--format=%(refname:short)',
            '--sort=committerdate'
        ]).split('\n')

    @property
    def remotes(self) -> typing.Iterable[str]:
        """Return all remotes of this repository."""
        return self.__run_git_command(['remote']).split('\n')

    def fetch(self) -> None:
        """Fetch updates from remote repositories."""
        self.__run_git_command(['fetch', '--all'])

    def __run_git_command(self, params: list) -> str:
        """Run a git command on this repository."""
        prefix = ['git', '-C', self.__path]
        return subprocess.check_output(prefix + params, stderr=subprocess.STDOUT).decode().strip()


@functools.total_ordering
class GitVersion(Version):
    """Git implementation of a VCS version."""

    def __init__(self, repository: GitRepository, rev: str):
        """Constructor.

        Args:
            repository: The parent Git repository.
            rev: Any git-parsable revision (e.g. a sha1-hash, branch name, or tag).

        Raises:
            InvalidInputError: When the given rev is not a git-parsable revision.
        """
        self.__repository = repository
        try:
            self.__sha1 = self.__run_git_command(['rev-parse', '--verify', rev + '^{commit}'])
            self.__symbolic_ref = self.__run_git_command(['rev-parse', '--symbolic-full-name', '--abbrev-ref', rev])
        except subprocess.CalledProcessError:
            raise InvalidInputError('{rev} is not a valid git revision'.format(rev=rev))

    @property
    def date(self) -> datetime.datetime:
        timestamp = self.__run_git_command(['show', '-s', '--format=%ct', self.id])
        return datetime.datetime.fromtimestamp(int(timestamp))

    @property
    def repository(self) -> GitRepository:
        return self.__repository

    @property
    def id(self) -> str:
        return self.__sha1

    @property
    def version_id(self) -> str:
        """PEP 440 compatible version identifier.

        Returns:
            - `<tag>` if the current version is tagged. Leading `v`'s are stripped.
            - `<tag>.post0.dev<N>+g<SHA1>.<branch>` if the current version is not tagged.
                Here <N> is the number of commits since <tag>.
            - `0.0.0.post0.dev<N>+g<SHA1>.<branch>` if the current version is not tagged
                and there are no tags in its ancestry.
        """
        try:
            # `git describe --tags` returns `<tag>-<N>-g<SHA1>` if we're not on a tag,
            # or just `<tag>` if we are on a tag. We need to massage the tail into a PEP 440 format.
            # We do this by stripping the tag and making some replacements.
            version_full = self.__run_git_command(['describe', '--tags', self.__sha1])
            version_tag = self.__run_git_command(['describe', '--tags', '--abbrev=0', self.__sha1])
            version_tail = version_full[len(version_tag):]
            # Make replacements and stitch prefix and postfix together.
            # Also remove leading `v`'s, which are common prefixes for git version tags.
            version = version_tag.lstrip('v') + version_tail.replace('-', '.post0.dev', 1).replace('-', '+', 1)
        except subprocess.CalledProcessError:
            # If we got here, there are probably no tags (as `git describe --tags` fails in that case).
            # Return a version with leading `0.0` tag.
            num_commits = self.__run_git_command(['rev-list', '--count', self.__sha1])
            short_sha1 = self.__run_git_command(['rev-parse', '--short', self.__sha1])
            version = '0.0.0.post0.dev{num_commits}+g{short_sha1}'.format(
                num_commits=num_commits,
                short_sha1=short_sha1
            )
        # Add symbolic reference iff there's a local identifier.
        if '+' in version and self.__symbolic_ref:
            # PEP 440 only allows [0-9a-zA-z\.] in the local identifier.
            replaced = re.sub('[^0-9a-zA-z]', '.', self.__symbolic_ref)
            version += '.{symbol}'.format(symbol=replaced)
        return version

    def newest(self, branch: str = None) -> 'GitVersion':

        self.repository.fetch()

        if branch is not None:
            pattern = r'(({remotes})\/)?{branch}'.format(
                remotes='|'.join(self.repository.remotes),
                branch=fnmatch.translate(branch)
            )
            regex = re.compile(pattern)

        versions = []
        for rep_branch in self.repository.branches:
            if branch is not None and not regex.match(rep_branch):
                continue
            version = GitVersion(self.repository, rep_branch)
            if not self.is_ancestor_of(version):
                continue
            versions.append(version)

        if versions:
            return max(versions)
        else:
            return self

    def is_ancestor_of(self, other: 'GitVersion') -> bool:
        """Check whether this revision is an ancestor of another revision."""
        if self.repository != other.repository:
            return False
        try:
            self.__run_git_command(['merge-base', '--is-ancestor', self.id, other.id])
            return True
        except subprocess.CalledProcessError:
            return False

    def last_change_on(self, filename: str) -> typing.Optional['GitVersion']:
        """Return the last version which changed the given file, looking back from this version.

        Returns:
            None if there is no such a version.
        """
        sha1 = self.__run_git_command(['log', '-n', '1', '--format=format:%H', self.id, '--', filename])
        if sha1:
            return GitVersion(self.repository, sha1)
        else:
            return None

    def __run_git_command(self, params: list) -> str:
        """Run a git command on the parent repository."""
        prefix = ['git', '-C', self.repository.path]
        return subprocess.check_output(prefix + params, stderr=subprocess.STDOUT).decode().strip()

    def __eq__(self, other: Version) -> bool:
        """Check equality via SHA1 hashes."""
        return self.id == other.id and self.repository == other.repository

    def __gt__(self, other: Version) -> bool:
        """Compare order via version dates."""
        if self.repository == other.repository:
            return self.date > other.date
        else:
            return NotImplemented
