"""Interfaces for VCS repositories and versions"""

import abc
import datetime
import typing


class Repository(abc.ABC):
    """Interface for VCS repositories."""

    @abc.abstractmethod
    def __init__(self, path: str) -> None:
        """Repository constructor.

        Raises:
            InvalidInputError: When the given path is not campatible with the repository type.
        """
        pass

    @property
    @abc.abstractmethod
    def path(self) -> str:
        """Return the path of this repository."""
        pass

    @classmethod
    @abc.abstractmethod
    def is_compatible_path(cls, path: str) -> bool:
        """Check whether the given path is compatible with this type of VCS repository."""
        pass

    @abc.abstractmethod
    def current_version(self) -> 'Version':
        """Return the current version of the repository."""
        pass

    def __eq__(self, other: 'Repository') -> bool:
        """Check equality via paths."""
        return self.path == other.path


class Version(abc.ABC):
    """Interface for versions of a VCS repository

    A version always describes the repository as a whole, not just a single file.
    """

    @property
    @abc.abstractmethod
    def id(self) -> str:
        """The unique identifier of this version."""
        pass

    @property
    @abc.abstractmethod
    def version_id(self) -> str:
        """PEP 440 compatible version identifier."""

    @property
    @abc.abstractmethod
    def date(self) -> datetime.datetime:
        """The date this version was created."""
        pass

    @property
    @abc.abstractmethod
    def repository(self) -> Repository:
        """The parent repository of this version."""
        pass

    @abc.abstractmethod
    def newest(self, branch: str = None) -> 'Version':
        """The newest version in the parent repository which has this one as an ancestor.

        Args:
            branch: If given, the newest version is restricted to be in the given branch / tag / label.
                Can be a glob pattern, so e.g. 'v1.*' will match 'v1.1', 'v1.2', etc.
        """
        pass

    @abc.abstractmethod
    def last_change_on(self, filename: str) -> typing.Optional['Version']:
        """Return the last version which changed the given file, looking back from this version.

        Returns:
            None if there is no such version.
        """
        pass
